import jpype
import uuid
import time
import pygeoip

from common.java import JavaInterface, Callback, logging
from dixi.models import NodeLocation

java = JavaInterface()

from django.db import models
from django.conf import settings
from django.template.defaultfilters import floatformat, filesizeformat

class MonitoringCallback(Callback):
    rule = None
    subscription = None
    
    def callbackInvoked(self, data):
        try:
            if self.subscription:
                if isinstance(data, (java.IntegerMetric, java.FloatMetric, java.DoubleMetric, java.StringMetric)):
                    logging.info('New notification: METRIC %s' % data.getName())
                    
                    value_type = Notification.VALUE_TYPE_NONE
                    
                    if isinstance(data, java.IntegerMetric):
                        value_type = Notification.VALUE_TYPE_INTEGER
                    elif isinstance(data, java.LongMetric):
                        value_type = Notification.VALUE_TYPE_LONG
                    elif isinstance(data, java.FloatMetric):
                        value_type = Notification.VALUE_TYPE_FLOAT
                    elif isinstance(data, java.DoubleMetric):
                        value_type = Notification.VALUE_TYPE_DOUBLE
                    elif isinstance(data, java.StringMetric):
                        value_type = Notification.VALUE_TYPE_STRING
                    
                    notification = Notification(
                        notification_type=Notification.NOTIFICATION_TYPE_METRIC,
                        name=data.getName(),
                        value=str(data.getValue()),
                        value_type=value_type,
                        unit=data.getValueUnit().name(),
                        vo=str(data.domains.vo),
                        user=str(data.domains.user),
                        user_group=str(data.domains.userGroup),
                        job=str(data.domains.job),
                        resource=str(data.domains.resource),
                        
                        subscription=self.subscription,
                        query=self.query,
                    )
                    
                    notification.save()
                    
                elif isinstance(data, java.Event):
                    logging.info('New notification: EVENT %s' % data.getName())
                    
                    notification = Notification(
                        notification_type=Notification.NOTIFICATION_TYPE_EVENT,
                        name=data.getName(),
                        subscription=self.subscription
                    )
                    
                    notification.save()
        except:
            logging.exception('New notification error')
            import traceback
            traceback.print_exc()

def authenticate(username, password):
    try:
        user = java.xvoms_util.getUser(username, password)
        return user.guid
    except jpype.JException(java.VOException):
        pass
    
    return None

gic = None

try:
    gic = pygeoip.GeoIP(settings.GEOIP_DB_PATH, pygeoip.MEMORY_CACHE)
except IOError:
    pass

def geoip(ip):
    # by ip
    record = gic.record_by_addr(ip) if gic else None
    
    # stored in db
    if not record:
        try:
            node = NodeLocation.objects.filter(node__startswith='://'+ip)[0]
            record = { 
                'latitude': str(node.lat),
                'longitude': str(node.lng)
            }
        except:
            pass
    
    # fallback
    if not record:
        record = settings.GEOIP_FALLBACK_DATA
    
    return record

class MonitoringManager(object):
    monitoring_session = None
    
    def __init__(self, monitoring_session):
        self.monitoring_session = monitoring_session
    
    def generate_rule(self, handle, vo=None, user=None, usergroup=None,
                        job=None, resource=None, aggregation_method=None,
                        aggregation_params=None, comparator=None,
                        condition=None):
        
        params = [
            ('vo', vo),
            ('user', user),
            ('usergroup', usergroup),
            ('job', job),
            ('resource', resource)
        ]
        
        params = [(x[0], x[1]) for x in params if x[1]]
        
        query = '@'
        
        if params:
            query += '{%s}' % ','.join(['%s=[%s]' % x for x in params])
        
        if aggregation_method:
            query += '%s(%s,%s)' % (aggregation_method, handle,
                                    ','.join([str(x) for x in aggregation_params]))
        else:
            query += handle
        
        if condition is not None:
            query += '%s%s' % (comparator, str(condition))
        
        rule = 'webmon_%s_%s' % (handle, uuid.uuid4())
        
        logging.info('Adding new notification rule %s.' % rule)
        
        java.XMonMng.addNotificationXATI(query, rule, java.certificate)
        
        return rule
    
    def get_vos(self):
        rs = []
        
        for item in java.xvoms_util.getAllVOs().toArray():
            rs.append({
                'name': item.name,
                'gvid': item.gvid,
                'description': item.description,
            })
        
        return rs
    
    def get_users(self):
        rs = []
        
        for item in java.xvoms_util.getAllUsers().toArray():
            rs.append({
                'realname': item.realname,
                'guid': item.guid,
                'username': item.username,
                'status': item.status,
                'affiliation': item.affiliation,
                'description': item.description,
            })
        
        return rs
    
    def get_resources(self):
        rs = []
        
        for item in java.XRCAServer.getRegisteredResources(java.certificate).toArray():
            ip = item.communicationAddress.externalAddress.hostAddress
            port = item.communicationAddress.port
            rs.append({
                'resource': '%s:%s' % (ip, port),
                'ip': ip,
                'port': port,
                'unique_id': item.hostUniqueID,
                'computation_resource': bool(item.isComputationResource),
                'operating_system_name': item.operatingSystemName,
                'processor_architecture': item.processorArchitecture,
                'ram_size': filesizeformat(item.RAMSize),
            })
        
        return rs
    
    def get_rules(self):
        rs = []
        
        hashmap = java.XMonMng.getMonitoringRulesXATI(java.certificate)
        
        for key in hashmap.keySet().toArray():
            rs.append({
                'rule': key,
                'query': hashmap.get(key),
            })
        
        return rs
        
    def monitoring_rules(self):
        rules = self.get_rules()
        
        subscriptions = self.monitoring_session.subscriptions()
        
        for rule in rules:
            rule['subscribed'] = False
            
            if subscriptions.filter(rule=rule['rule']):
                rule['subscribed'] = True
        
        return rules
    
    def monitoring_subscribe(self, rule):
        logging.info('Subscribing to rule %s.' % rule)
        
        callback = MonitoringCallback()
        proxy = jpype.JProxy([java.ICallbackListener, java.Serializable], inst=callback)
        subscription_id = java.XMonMng.subscribeXATI(rule, java.XATI.setCallback(proxy), java.certificate)
        
        subscription = Subscription.objects.create(
            id=subscription_id,
            rule=rule,
            monitoring_session=self.monitoring_session
        )
        
        hashmap = java.XMonMng.getMonitoringRulesXATI(java.certificate)
        
        callback.subscription = subscription
        callback.query = hashmap.get(rule)
        
        return subscription_id
    
    def monitoring_unsubscribe(self, subscription_id):
        logging.info('Unsubscribing from rule with subscription ID %s.' % subscription_id)
        java.XMonMng.unsubscribeXATI(subscription_id, java.certificate)

class MonitoringSessionManager(models.Manager):
    def get_monitoring_session(self, request):
        monitoring_session = request.session.get('monitoring_session', None)
        
        if monitoring_session:
            monitoring_session.save()
        else:
            user = request.session['user']
            monitoring_session = MonitoringSession.objects.create(user=user)
            request.session['monitoring_session'] = monitoring_session
            logging.info('Creating new monitoring session')
        
        return monitoring_session

class MonitoringSession(models.Model):
    user = models.CharField(max_length=36)
    created = models.DateTimeField(auto_now_add=True)
    last_access = models.DateTimeField(auto_now=True)
    
    objects = MonitoringSessionManager()
    
    class Meta:
        ordering = ('created',)
    
    def __unicode__(self):
        return u'%s @ %s' %  (self.user, self.created)
    
    def get_manager(self):
        if not '_monitoring_manager' in self.__dict__:
            self._monitoring_manager = MonitoringManager(self)
        
        return self._monitoring_manager
    
    def subscriptions(self):
        return self.subscription_set.all()
    
    def notifications(self):
        return Notification.objects.filter(subscription__monitoring_session=self)
    
    def pop_notifications(self):
        qs = Notification.objects.filter(subscription__monitoring_session=self)
        qs = qs.order_by('-id')
        
        notifications = list(qs)
        
        if notifications:
            qs.filter(id__lte=notifications[0].id).delete()
        
        return notifications
    
    def serialize(self):
        return {
            'id': self.id,
            'user': unicode(self.user),
            'created': unicode(self.created),
            'last_access': unicode(self.last_access)
        }

class Subscription(models.Model):
    id = models.CharField(max_length=36, primary_key=True)
    rule = models.CharField(max_length=255)
    monitoring_session = models.ForeignKey(MonitoringSession)
    created = models.DateTimeField(auto_now=True)
    
    class Meta:
        ordering = ('created',)
    
    def __unicode__(self):
        return u'%s for %s' %  (self.id, self.rule)
    
    def serialize(self):
        return {
            'id': self.id,
            'rule': self.rule,
            'monitoring_session': self.serializable_value('monitoring_session')
        }

class Notification(models.Model):
    NOTIFICATION_TYPE_METRIC = 'METRIC'
    NOTIFICATION_TYPE_EVENT = 'EVENT'
    
    NOTIFICATION_TYPES = (
        (NOTIFICATION_TYPE_METRIC, 'Metric'),
        (NOTIFICATION_TYPE_EVENT, 'Event'),
    )
    
    VALUE_TYPE_NONE = 'NONE'
    VALUE_TYPE_INTEGER = 'INTEGER'
    VALUE_TYPE_LONG = 'LONG'
    VALUE_TYPE_FLOAT = 'FLOAT'
    VALUE_TYPE_DOUBLE = 'DOUBLE'
    VALUE_TYPE_STRING = 'STRING'
    
    VALUE_TYPES = (
        (VALUE_TYPE_NONE, 'None'),
        (VALUE_TYPE_INTEGER, 'Integer'),
        (VALUE_TYPE_LONG, 'Long'),
        (VALUE_TYPE_FLOAT, 'Float'),
        (VALUE_TYPE_DOUBLE, 'Double'),
        (VALUE_TYPE_STRING, 'String'),
    )
    
    notification_type = models.CharField(max_length=50, choices=NOTIFICATION_TYPES, default=NOTIFICATION_TYPE_METRIC)
    name = models.CharField(max_length=255)
    value = models.TextField(null=True, blank=True)
    value_type = models.CharField(max_length=50, choices=VALUE_TYPES, default=VALUE_TYPE_NONE)
    unit = models.CharField(max_length=50, null=True, blank=True)
    vo = models.CharField(max_length=50, null=True, blank=True)
    user = models.CharField(max_length=50, null=True, blank=True)
    user_group = models.CharField(max_length=50, null=True, blank=True)
    job = models.CharField(max_length=50, null=True, blank=True)
    resource = models.CharField(max_length=50, null=True, blank=True)
    
    subscription = models.ForeignKey(Subscription)
    query = models.TextField(null=True, blank=True)
    datetime = models.DateTimeField(auto_now_add=True)
    
    class Meta:
        ordering = ('-id',)
    
    def __unicode__(self):
        return u'%s: %s %s %s @ %s' %  (self.get_notification_type_display(), self.name, self.value, self.unit, self.datetime)
    
    def get_value(self):
        unit = (' %s' % (self.unit)) if self.unit else ''
        
        if self.unit == 'PERCENT':
            unit = '%'
            
        if self.value_type in (self.VALUE_TYPE_INTEGER, self.VALUE_TYPE_LONG):
            return '%d%s' % (self.value, unit)
        
        elif self.value_type == self.VALUE_TYPE_FLOAT:
            return '%s%s' % (floatformat(self.value, -2), unit)
        
        elif self.value_type == self.VALUE_TYPE_DOUBLE:
            return '%s%s' % (floatformat(self.value, -2), unit)
        
        elif self.value_type == self.VALUE_TYPE_STRING:
            return '%s%s' % (self.value, unit)
        
        return unit
    
    def original_value(self):
        if self.value_type == self.VALUE_TYPE_INTEGER:
            return int(self.value)
        elif self.value_type == self.VALUE_TYPE_LONG:
            return long(self.value)
        elif self.value_type == self.VALUE_TYPE_FLOAT:
            return float(self.value)
        elif self.value_type == self.VALUE_TYPE_DOUBLE:
            return float(self.value)
        elif self.value_type == self.VALUE_TYPE_STRING:
            return self.value
    
    def serialize(self):
        try:
            import random
            
            data = {
                'id': self.id,
                'notification_type': self.get_notification_type_display(),
                'name': self.name,
                'value': self.get_value(),
                'value_original': self.original_value(),
                'value_unit': self.unit,
                'vo': self.vo,
                'user': self.user,
                'user_group': self.user_group,
                'job': self.job,
                'resource': self.resource,
                'subscription': self.serializable_value('subscription'),
                'query': self.query,
                'timestamp': self.datetime.isoformat(),
                'latitude': None,
                'longitude': None
            }
            
            resource = self.resource
            
            if resource:
                ip = resource[:resource.find(':')] if ':' in resource else resource
                data.update(geoip(ip) or {})
            
            return data
        except:
            import traceback
            traceback.print_exc()
            return {}

if False:
    from monitoring.models import *

    monitoring_session = MonitoringSession.objects.latest('id')
    monitoring_manager = monitoring_session.get_manager()
    java.xvoms_util
    
    java.xvoms_util.getAllVOs()
    
    id = monitoring_manager.subscribe('cpu_user')
    monitoring_manager.unsubscribe(id)
    
    id1 = monitoring_manager.subscribe('disk_free')
    monitoring_manager.unsubscribe(id1)
    
    id2 = monitoring_manager.subscribe('job_started')
    monitoring_manager.unsubscribe(id2)
    